import sys
import random
import pyglet
import const
import util


class GameWindow(pyglet.window.Window):

    def __init__(self):
        super(GameWindow, self).__init__(const.WIDTH, const.HEIGHT,
                                         'People and Planes')
        self.start_level(0)

    def start_level(self, level):
        self.current_window = Game(self, level)


class Game:

    def __init__(self, window, level):
        self.window = window
        self.level_n = level
        self.level = Level(*const.LEVELS[level])

        self.finished = 0
        self.paused = False
       
        self.batch = pyglet.graphics.Batch()

        self.bg = pyglet.sprite.Sprite(pyglet.resource.image('bg.png'),
                                       0, 0, batch=self.batch,
                                       group=const.BG_GROUP)
        
        self.timer = Timer(self.level.init_timer, self.level_n, self.batch)
        self.score = Score(self.level.target_score, self.batch)

        init_pos = [256, 400]
        self.cloud = Cloud(init_pos[0], init_pos[1], self.batch)
        self.string = String(init_pos[0], init_pos[1], self.level.length,
                             self.batch, self.cloud)
        self.cloud.set_string(self.string)
        self.person_gen = PersonGenerator(self.batch, self.string,
                                          self.level.max_people,
                                          self.level.n_colors)
        self.plane_gen = PlaneGenerator(self.batch, self.person_gen,
                                        self.cloud, self.level.n_colors,
                                        self.score)

        self.window.push_handlers(self.on_draw, self.on_key_press,
                                  self.on_key_release)
        pyglet.clock.schedule_interval(self.update, 1.0/const.FPS)

    def finish(self):
        self.window.pop_handlers()
        pyglet.clock.unschedule(self.update)

    def on_draw(self):
        self.window.clear()
        self.batch.draw()
 
    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.LEFT:
            self.cloud.move_left = True
        elif symbol == pyglet.window.key.RIGHT:
            self.cloud.move_right = True
        elif symbol == pyglet.window.key.UP:
            self.cloud.move_up = True
        elif symbol == pyglet.window.key.DOWN:
            self.cloud.move_down = True
        elif symbol == pyglet.window.key.R:
            self.end_level(False)
        elif symbol == pyglet.window.key.P:
            if not self.paused: self.pause()
            else: self.unpause()

    def on_key_release(self, symbol, modifiers):
        if symbol == pyglet.window.key.LEFT:
            self.cloud.move_left = False
        elif symbol == pyglet.window.key.RIGHT:
            self.cloud.move_right = False
        elif symbol == pyglet.window.key.UP:
            self.cloud.move_up = False
        elif symbol == pyglet.window.key.DOWN:
            self.cloud.move_down = False

    def pause(self):
        self.paused = True
        self.paused_label = pyglet.text.Label('Paused', batch=self.batch,
                                              x=const.WIDTH / 2,
                                              y=const.HEIGHT / 2,
                                              font_size=36,
                                              anchor_x = 'center',
                                              anchor_y = 'center')

    def unpause(self):
        self.paused = False
        self.paused_label.delete()
        self.paused_label = None

    def update(self, dt):
        if self.paused:
            return
        self.cloud.update(dt)
        self.string.update(dt)
        self.person_gen.update(dt)
        self.plane_gen.update(dt)
        self.timer.update(dt)
        self.score.update()
        if self.timer.done:
            self.end_level(False)
        if self.score.done:
            self.end_level(True)
        if self.finished == 1:
            self.really_win()
        if self.finished == -1:
            self.really_lose()

    def end_level(self, win):
        if win:
            if self.level_n == const.N_LEVELS - 1:
                text = 'Congratulations! You win!'
            else:
                text = 'Level finished!'
            pyglet.clock.schedule_once(self.win, 2)
        else:
            text = 'Time out!'
            pyglet.clock.schedule_once(self.lose, 2)
        self.label = pyglet.text.Label(text, batch=self.batch,
                                       x=const.WIDTH / 2,
                                       y=const.HEIGHT / 2, font_size=36,
                                       anchor_x = 'center',
                                       anchor_y = 'center')

    def win(self, dt):
        self.finished = 1

    def lose(self, dt):
        self.finished = -1

    def really_win(self):
        self.finish()
        self.delete()
        if self.level_n == const.N_LEVELS - 1:
            sys.exit()
        else:
            self.window.start_level(self.level_n + 1)
        self.finished = 0

    def really_lose(self):
        self.finish()
        self.delete()
        self.window.start_level(self.level_n)
        self.finished = 0

    def delete(self):
        self.batch = None
        self.bg.delete()
        del(self.timer)
        self.timer = None
        del(self.score)
        self.score = None
        del(self.cloud)
        self.cloud = None
        del(self.string)
        self.string = None
        del(self.person_gen)
        self.person_gen = None
        del(self.plane_gen)
        self.plane_gen = None


class String:

    def __init__(self, x, y, length, batch, cloud):
        self.offset_x = 64
        self.offset_y = 2
        x += self.offset_x
        y += self.offset_y
        self.length = length
        self.batch = batch
        self.cloud = cloud
        self.points = []
        self.points_vel = []
        self.dist = 5
        for i in range(self.length):
            self.points.append([x, y - i * self.dist])
            self.points_vel.append([0.0, 0.0])
        self.vlist = self.batch.add(self.length, pyglet.gl.GL_LINE_STRIP,
                                    const.STRING_GROUP, 'v2f')

    def update_points(self):
        for i in range(1, self.length):
            p = self.points[i]
            prev = self.points[i-1]
            if i == 1: prev2 = prev
            else: prev2 = self.points[i-2]
            if i <= 2: prev3 = prev2
            else: prev3 = self.points[i-3]
            self.points_vel[i][0] += (prev[0] - p[0]) * 0.3
            self.points_vel[i][0] += (prev2[0] - p[0]) * 0.3
            self.points_vel[i][0] += (prev3[0] - p[0]) * 0.2
            self.points_vel[i][0] *= 0.4

    def update_vlist(self):
        vertices = []
        for p in self.points:
            vertices += p
        self.vlist.vertices = vertices

    def update_vel(self, dt):
        for i in range(self.length):
            self.points[i][0] += self.points_vel[i][0]# * 30 * dt
            self.points[i][1] += self.points_vel[i][1]# * 30 * dt

    def update(self, dt):
        self.points[0][0] = self.cloud.x + self.offset_x
        self.points[0][1] = self.cloud.y + self.offset_y
        for i in range(1, self.length):
            self.points[i][1] = self.points[i-1][1] - self.dist
        self.update_points()
        self.update_vel(dt)
        self.update_vlist()


class Cloud(pyglet.sprite.Sprite):

    def __init__(self, x, y, batch):
        self.image_normal = pyglet.resource.image('cloud.png')
        self.image_hurt = pyglet.resource.image('cloud_hurt.png')
        super(Cloud, self).__init__(self.image_normal, x=x, y=y, batch=batch,
                                    group=const.CLOUD_GROUP)
        self.speed = 150
        self.move_left = False
        self.move_right = False
        self.move_up = False
        self.move_down = False
        self.stunned = False
        self.stunned_timer = const.FPS

    def set_string(self, string):
        self.string = string

    def stun(self):
        self.stunned = True
        self.image = self.image_hurt
        self.stunned_timer = const.FPS
        
    def unstun(self):
        self.stunned = False
        self.image = self.image_normal

    def update(self, dt):
        if self.stunned:
            self.stunned_timer -= 1
            if self.stunned_timer <= 0:
                self.unstun()
            return
        if self.move_right:
            self.x += self.speed * dt
        elif self.move_left:
            self.x -= self.speed * dt
        if self.move_up:
            self.y += self.speed * dt
        elif self.move_down:
            self.y -= self.speed * dt
        if self.x + 64 > const.WIDTH:
            self.x = const.WIDTH - 64
        if self.x + 64 < 0:
            self.x = -64
        if self.y + 2 - self.string.length * self.string.dist > const.HEIGHT:
            self.y = const.HEIGHT - 2 + self.string.length * self.string.dist
        if self.y + 2 < 0:
            self.y = -2            


class Person(pyglet.sprite.Sprite):

    def __init__(self, x, y, batch, string, color):
        self.string = string
        full_image = pyglet.resource.image('person.png')
        self.image_seq = pyglet.image.ImageGrid(full_image, 1, 4)

        waving_rate = 0.2
        self.anim_waving = pyglet.image.Animation.from_image_sequence(
            self.image_seq[0:2], waving_rate, True)
        self.grabbing_left = self.image_seq[2]
        self.grabbing_right = self.image_seq[3]

        super(Person, self).__init__(self.anim_waving, x=x, y=y, batch=batch,
                                     group=const.PERSON_GROUP)
        self.color = color
        self.grabbed = False
        self.grab_index = -1
        self.need_to_restore_pos = False
        self.original_pos = [x, y]
        self.destroyed = False

    def check_string_grab(self):
        for i in range(self.string.length):
            p = self.string.points[i]
            if util.hit(self, p[0], p[1]):
                self.grab(i)
                break

    def grab(self, i):
        self.grabbed = True
        self.grab_index = i
        if (self.x + self.image_seq[0].width/2.0 <
            self.string.points[self.grab_index][0]):
            self.image = self.grabbing_left
        else:
            self.image = self.grabbing_right
        self.need_to_restore_pos = True

    def update(self, dt):
        if not self.grabbed:
            self.check_string_grab()
        else:
            p = self.string.points[self.grab_index]
            if self.image == self.grabbing_left:
                self.x = p[0] - 16
                self.y = p[1] - 20
            else:
                self.x = p[0] - 4
                self.y = p[1] - 20                
        

class PersonGenerator:

    def __init__(self, batch, string, max_people, n_colors):
        self.people = []
        self.batch = batch
        self.string = string
        self.spawn_pos = const.SPAWN_POS[:]
        self.used_pos = []
        self.max_people = max_people
        self.colors = const.COLORS[:n_colors]

    def generate(self):
        pos = random.choice(self.spawn_pos)
        person = Person(pos[0], pos[1], self.batch, self.string,
                        random.choice(self.colors))
        self.people.append(person)
        self.spawn_pos.remove(pos)
        self.used_pos.append(pos)

    def restore_pos(self, pos):
        self.used_pos.remove(pos)
        self.spawn_pos.append(pos)

    def update(self, dt):
        if len(self.people) < self.max_people:
            if random.randint(1, 30) == 1:
                self.generate()
        for person in self.people:
            person.update(dt)
            if person.need_to_restore_pos:
                self.restore_pos(person.original_pos)
                person.need_to_restore_pos = False
            if person.destroyed:
                self.people.remove(person)
                person.delete()


class Plane(pyglet.sprite.Sprite):

    def __init__(self, x, y, batch, color, direction):
        if direction == 1:
            full_image = pyglet.resource.image('plane_right.png')
        else:
            full_image = pyglet.resource.image('plane_left.png')            
        self.image_seq = pyglet.image.ImageGrid(full_image, 1, 2)
        
        self.anim_plane = pyglet.image.Animation.from_image_sequence(
            self.image_seq[0:2], 0.05, True)
        
        super(Plane, self).__init__(self.anim_plane, x=x, y=y, batch=batch,
                                    group=const.PLANE_GROUP)
        self.color = color
        self.speed = 120
        self.direction = direction
        self.destroyed = False

    def update(self, dt):
        self.x += self.direction * self.speed * dt
        if ((self.direction == 1 and self.x >= const.WIDTH) or
            (self.direction == -1 and self.x + self.width < 0)):
            self.destroyed = True


class PlaneGenerator:

    def __init__(self, batch, person_gen, cloud, n_colors, score):
        self.batch = batch
        self.person_gen = person_gen
        self.cloud = cloud
        self.planes = []
        self.signals = []
        self.range = [120, 444]
        self.start_delay = 30
        self.delay = self.start_delay
        self.chance = 20
        self.colors = const.COLORS[:n_colors]
        self.score = score

    def generate(self):
        color = random.choice(self.colors)
        if random.randint(0, 1) == 0:
            pos = [-146, 0]
            direction = 1
        else:
            pos = [const.WIDTH + 146, 0]
            direction = -1
        pos[1] = random.randint(*self.range)
        plane = Plane(pos[0], pos[1], self.batch, color, direction)
        self.planes.append(plane)

    def generate_signal(self, x, y, color, right):
        signal = Signal(x, y, self.batch, color, right)
        self.signals.append(signal)

    def check_collision(self):
        for person in self.person_gen.people:
            if person.grabbed and not person.destroyed:
                for plane in self.planes:
                    if util.hit_sprite(person, plane):
                        person.destroyed = True
                        right_color = (person.color == plane.color)
                        self.generate_signal(plane.x + 40, plane.y + 20,
                                             person.color, right_color)
                        if right_color:
                            self.score.add_score(1)
                        else:
                            self.score.add_score(-1)
        for plane in self.planes:
            if self.cloud.stunned == False:
                if (util.hit_rect_sprite(self.cloud.x + 43, self.cloud.x + 70,
                                         self.cloud.y + 56, self.cloud.y + 33,
                                         plane) or
                    util.hit_rect_sprite(self.cloud.x + 59, self.cloud.x + 73,
                                         self.cloud.y + 13, self.cloud.y + 0,
                                         plane)):
                    self.cloud.stun()
        
    def update(self, dt):
        for plane in self.planes:
            plane.update(dt)
            if plane.destroyed:
                self.planes.remove(plane)
                plane.delete()
        for signal in self.signals:
            signal.update(dt)
            if signal.destroyed:
                self.signals.remove(signal)
                signal.delete()
        self.check_collision()
        self.delay -= 1
        if self.delay <= 0:
            if random.randint(1, self.chance) == 1:
                self.generate()
                self.delay = self.start_delay


class Signal(pyglet.sprite.Sprite):

    def __init__(self, x, y, batch, color, good):
        if good:
            image = pyglet.resource.image('plus.png')
        else:
            image = pyglet.resource.image('minus.png')
        super(Signal, self).__init__(image, x, y, batch=batch,
                                     group=const.SIGNAL_GROUP)
        self.color = color
        self.timer = 0.5 * const.FPS
        self.destroyed = False

    def update(self, dt):
        self.y += 20 * dt
        self.timer -= 1
        if self.timer <= 0:
            self.destroyed = True


class Timer:

    def __init__(self, timer, level, batch):
        self.timer = float(timer)
        self.level = level
        text = '%02d:%02d [Level %d of %d]' % (self.timer / 60,
                                               self.timer % 60,
                                               self.level + 1,
                                               const.N_LEVELS)
        self.label = pyglet.text.Label(text, batch=batch, x=10,
                                       y=const.HEIGHT - 20)
        self.done = False

    def update(self, dt):
        self.timer -= dt
        self.update_label()
        if self.timer <= 0:
            self.done = True

    def update_label(self):
        self.label.text = '%02d:%02d [Level %d of %d]' % (self.timer / 60,
                                                          self.timer % 60,
                                                          self.level + 1,
                                                          const.N_LEVELS)


class Score:

    def __init__(self, target_score, batch):
        self.score = 0
        self.target_score = target_score
        text = 'Score: %2d / %2d [target]' % (self.score, self.target_score)
        self.label = pyglet.text.Label(text, batch=batch, x=const.WIDTH-180,
                                       y=const.HEIGHT-20)
        self.done = False

    def add_score(self, score):
        self.score += score
        if self.score < 0:
            self.score = 0
        if self.score >= self.target_score:
            self.done = True

    def update(self):
        self.label.text = 'Score: %2d / %2d [target]' % (self.score,
                                                         self.target_score)


class Level:

    def __init__(self, number, init_timer, max_people, n_colors, length,
                 target_score):
        self.number = number
        self.init_timer = init_timer
        self.max_people = max_people
        self.n_colors = n_colors
        self.length = length
        self.target_score = target_score
